package main

import (
	"bytes"
	"encoding/json"
	"fmt"
	"io/ioutil"
	"log"
	"net/http"
	"strconv"
	"time"

	"mfs.de/spacedata"
)

func main() {
	// Tracking. :)
	start := time.Now()

	// Hier listen wir alle Crew Mitglieder der ISS auf.
	providePeopleOnISS()

	// Hier kommt die aktuelle Position der ISS hin - die Daten brauchen wir gleich noch.
	lat, long := provideCurrentPosition()

	// Eine kleine Auflistung der nächsten Aufenthalte unserer ISS an diesem Ort.
	provideEstimatesRisings(lat, long)

	// Auswertungskrams
	duration := time.Now().Sub(start)
	fmt.Printf("Das waren jetzt %0.2f Sekunden an Computerarbeitszeit.\nSchönen Tag noch. :)\n", duration.Seconds())
}

func providePeopleOnISS() {
	var peopleProvider spacedata.Providing = spacedata.PeopleProvider{}
	serverUriString := peopleProvider.EndpointString()
	logUri(serverUriString)

	httpClient := http.DefaultClient
	response, err := httpClient.Get(serverUriString)
	if err != nil {
		log.Fatalf("Darn, Dude! I did not expect the Astronauts request uri string %#v to be invalid, but it threw: %v", serverUriString, err)
	}

	jsonData, err := ioutil.ReadAll(response.Body)
	if err != nil {
		log.Fatalf("Darn, Dude! I did not expect the Astronauts data read to fail, but it did with error: %v", err)
	}

	result, err := peopleProvider.ParseResponse(jsonData)
	if err != nil {
		log.Fatalf("Obviously we failed to parse the json data %#v\n with error: %v", string(jsonData), err)
	}
	if len(result) < 1 {
		log.Fatal("Seems like result is empty, but it should not be!")
	}

	peopleInSpace := make(map[string][]string, 0)
	reader := bytes.NewReader(result)
	decoder := json.NewDecoder(reader)
	err = decoder.Decode(&peopleInSpace)
	if err != nil {
		log.Fatalf("decoding people in space returned error: %v", err)
	}

	fmt.Println("Aktuell an Board der ISS sind:")

	for key, names := range peopleInSpace {
		if key == "ISS" {
			for idx, name := range names {
				fmt.Printf("%d. %v\n", idx+1, name)
			}
		}
	}
	fmt.Printf("\n--\n")
}

func logUri(uri string) {
	fmt.Printf("Hole Daten vom Server:\n%#v\n\t...\n", uri)
}

func provideEstimatesRisings(lat float64, long float64) {
	var passTimeProvider spacedata.Providing = spacedata.PassTimeProvider{}
	serverUriString := passTimeProvider.EndpointString()
	serverUriString += "?lat=" + strconv.FormatFloat(lat, 'f', 6, 64)
	serverUriString += "&lon=" + strconv.FormatFloat(long, 'f', 6, 64)
	logUri(serverUriString)

	httpClient := http.DefaultClient
	response, err := httpClient.Get(serverUriString)
	if err != nil {
		log.Fatalf("Darn, Dude! I did not expect the ISS position request uri string %#v to be invalid, but it threw: %v", serverUriString, err)
	}

	jsonData, err := ioutil.ReadAll(response.Body)
	if err != nil {
		log.Fatalf("Darn, Dude! I did not expect the ISS position data read to fail, but it did with error: %v", err)
	}
	result, err := passTimeProvider.ParseResponse(jsonData)
	if err != nil {
		log.Fatalf("Obviously we failed to parse the json data %#v\n with error: %v", string(jsonData), err)
	}
	if len(result) < 1 {
		log.Fatal("Seems like result is empty, but it should not be!")
	}

	fmt.Println("Die ISS schaut vermutlich wieder vorbei:")

	risingTimes := make([]spacedata.Risings, 0)
	reader := bytes.NewReader(result)
	decoder := json.NewDecoder(reader)
	err = decoder.Decode(&risingTimes)
	if err != nil {
		log.Fatalf("decoding estimated risings returned error: %v", err)
	}

	for _, rise := range risingTimes {
		fmt.Printf("Ab %v für etwa %d Sekunden\n", rise.IsoDateString(), rise.Duration)
	}

	fmt.Printf("\n--\n")
}

func provideCurrentPosition() (float64, float64) {
	var locationProvider spacedata.Providing = spacedata.LocationProvider{}
	serverUriString := locationProvider.EndpointString()
	logUri(serverUriString)

	httpClient := http.DefaultClient
	response, err := httpClient.Get(serverUriString)
	if err != nil {
		log.Fatalf("Darn, Dude! I did not expect the ISS position request uri string %#v to be invalid, but it threw: %v", serverUriString, err)
	}

	jsonData, err := ioutil.ReadAll(response.Body)
	if err != nil {
		log.Fatalf("Darn, Dude! I did not expect the ISS position data read to fail, but it did with error: %v", err)
	}

	result, err := locationProvider.ParseResponse(jsonData)
	if err != nil {
		log.Fatalf("Obviously we failed to parse the json data %#v\n with error: %v", string(jsonData), err)
	}
	if len(result) < 1 {
		log.Fatal("Seems like result is empty, but it should not be!")
	}

	var tempCoordinate spacedata.Coordinate
	reader := bytes.NewReader(result)
	decoder := json.NewDecoder(reader)
	err = decoder.Decode(&tempCoordinate)
	if err != nil {
		log.Fatalf("decoding iss coordinate returned error: %v", err)
	}
	issCoordinate := tempCoordinate.GpsCoordinate()

	fmt.Println("Die ISS befindet sich aktuell ungefähr hier:")
	fmt.Printf("Längengrad: %v; Breitengrad: %v\n", issCoordinate.Longitude(), issCoordinate.Latitude())
	fmt.Printf("Weitere Informationen liefert Google Maps unter \nhttps://www.google.com/maps/search/%v,%v", issCoordinate.Latitude(), issCoordinate.Longitude())
	fmt.Printf("\n\n--\n")

	return issCoordinate.Latitude(), issCoordinate.Longitude()
}
