//
//  ViewController.m
//  ISS Locator
//
//  Created by Hackathon on 28.09.19.
//  Copyright © 2019 Hackathon. All rights reserved.
//

#import "ViewController.h"
#import <Spacedata/Spacedata.h>

@interface ViewController()
@property(weak, atomic) IBOutlet MKMapView * mapView;

@property(strong, atomic) MKPointAnnotation * issAnnotation;
@property(strong, atomic) MKPolyline * issHistoricalOverlay;
@property(strong, atomic) CLLocationManager * locationManager;

@property(strong, atomic) UILabel * peopleInfo;
@property(copy, nonatomic) NSString * multilinePeopleNames;
@property(strong, atomic) UILabel * risingInfo;
@property(copy, nonatomic) NSString * multilineRisingTimes;

@property(strong, atomic) NSTimer * issLocationUpdateTimer;
@end

@implementation ViewController

#pragma mark View Life Cycle

- (void)viewDidLoad
{
    [super viewDidLoad];
    [[self mapView] setShowsUserLocation:YES];
    [[self mapView] setDelegate:self];
    [self updateIssMarker:self];
    
    NSTimer * apiRequestTimer = [NSTimer timerWithTimeInterval:3.14159 repeats:YES block:^(NSTimer * _Nonnull timer) {
        [self updateIssMarker:nil];
    }];
    [self setIssLocationUpdateTimer:apiRequestTimer];
    
    [self setLocationManager:[CLLocationManager new]];
    [[self locationManager] requestWhenInUseAuthorization];
    [[self locationManager] setDesiredAccuracy:kCLLocationAccuracyKilometer];
    [[self locationManager] setDistanceFilter:kCLDistanceFilterNone];
    [[self locationManager] setDelegate:self];
}

- (void)viewDidAppear:(BOOL)animated {
    [[self locationManager] startUpdatingLocation];
    [[NSRunLoop currentRunLoop] addTimer:[self issLocationUpdateTimer] forMode:NSDefaultRunLoopMode];
}

- (void)viewWillDisappear:(BOOL)animated {
    [[self issLocationUpdateTimer] invalidate];
    [[self locationManager] stopUpdatingLocation];
}

#pragma mark - Go—ObjC Glue Code

- (void) updateIssMarker:(id)sender
{
    SpacedataLocationProvider * provider = [SpacedataLocationProvider new];
    NSString * endpointString = [provider endpointString];
    
    NSURL * requestURL = [NSURL URLWithString:endpointString];
    NSURLRequest * apiRequest = [NSURLRequest requestWithURL:requestURL];
    
    [self jsonDataFromApiRequest:apiRequest provider:provider completion:^(id dataStruct) {
        dispatch_async(dispatch_get_main_queue(), ^{
            [self updateIssMarkerWithCoordinatesOfDTO:dataStruct moveToCoordinates:(sender == self)];
        });
    }];
}

- (void)jsonDataFromApiRequest:(NSURLRequest*)apiRequest
                      provider:(id<SpacedataProviding>)provider
                    completion:(void (^)(id))handler
{
    NSURLSession * apiSession = [NSURLSession sharedSession];
    NSURLSessionDataTask * locationTask = [apiSession dataTaskWithRequest:apiRequest completionHandler:^(NSData * _Nullable data, NSURLResponse * _Nullable response, NSError * _Nullable error) {
        if (error == nil) {
            NSData * result = [provider parseResponse:data error:&error];
            if (error == nil) {
                NSDictionary * dataStruct = [NSJSONSerialization JSONObjectWithData:result options:NSJSONReadingAllowFragments error:&error];
                if (error == nil) {
                    handler(dataStruct);
                }
            }
        }
    }];
    [locationTask resume];
}

- (void) updateIssMarkerWithCoordinatesOfDTO:(id)dto
                           moveToCoordinates:(BOOL)adjustCenter
{
    CLLocationCoordinate2D issCoordinate = CLLocationCoordinate2DMake([[dto objectForKey:@"latitude"] floatValue], [[dto objectForKey:@"longitude"] floatValue]);
    
    if ([self issAnnotation] == nil) {
        [self setIssAnnotation:[[MKPointAnnotation alloc] initWithCoordinate:issCoordinate title:@"Current ISS position" subtitle:nil]];
        [[self issAnnotation] setAccessibilityLabel:@"Position of ISS"];
        [self peopleOnIssForAnnotation:[self issAnnotation]];
        [[self mapView] addAnnotation:[self issAnnotation]];
    }
    else {
        MKPolyline * historicalOverlay = [self issHistoricalOverlay];
        
        NSUInteger pointsInOverlay = [historicalOverlay pointCount];
        CLLocationCoordinate2D coordinates[pointsInOverlay+1];

        if(historicalOverlay == nil) {
            historicalOverlay = [MKPolyline new];
        } else {
            [historicalOverlay getCoordinates:coordinates range:NSMakeRange(0, pointsInOverlay)];
        }
        coordinates[pointsInOverlay] = issCoordinate;
        
        [[self mapView] removeOverlay:historicalOverlay];
        
        MKPolyline * updatedOverlay = [MKPolyline polylineWithCoordinates:coordinates count:(pointsInOverlay +1)];
        
        [[self mapView] addOverlay:updatedOverlay];
        
        [self setIssHistoricalOverlay:updatedOverlay];
        [[self issAnnotation] setCoordinate:issCoordinate];
    }
    
    if(adjustCenter) {
        MKCoordinateRegion issRegion = MKCoordinateRegionMakeWithDistance(issCoordinate, 5000000, 5000000);
        [[self mapView] setRegion:issRegion];
    }
}

- (void)peopleOnIssForAnnotation:(id<MKAnnotation>)annotation
{
    SpacedataPeopleProvider * provider = [SpacedataPeopleProvider new];
    NSString * endpointString = [provider endpointString];
    
    NSString *documentsPath = [[NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES) objectAtIndex:0] stringByAppendingString:@"/"];
    
    [provider setDatabaseDirectory:documentsPath];
    NSURL * requestURL = [NSURL URLWithString:endpointString];
    NSURLRequest * apiRequest = [NSURLRequest requestWithURL:requestURL];
    
    [self jsonDataFromApiRequest:apiRequest provider:provider completion:^(id dataStruct) {
        dispatch_async(dispatch_get_main_queue(), ^{
            [self setPeopleFromJson:dataStruct];
        });
    }];
}

- (void) setPeopleFromJson:(id)json
{
    NSMutableString * people = [NSMutableString string];
    for (NSString *person in [json valueForKey:@"ISS"]) {
        if([people length] > 0) {
            [people appendString:@"\n"];
        }
        [people appendFormat:@"• %@",person];
    }
    [[self issAnnotation] setTitle:[NSString stringWithFormat:@"%lu People on ISS:", [[json valueForKey:@"ISS"] count]]];

    [self setMultilinePeopleNames:people];
}

- (void)setMultilinePeopleNames:(NSString *)multilinePeopleNames
{
    if(_multilinePeopleNames != multilinePeopleNames) {
        _multilinePeopleNames = [multilinePeopleNames copy];
        [self addAstronautDetailsToCalloutAccessory];
    }
}

- (void)addAstronautDetailsToCalloutAccessory
{
    if([[self multilinePeopleNames] length] > 0 && [self peopleInfo] != nil) {
        [[self peopleInfo] setText:[self multilinePeopleNames]];
    }
}

#pragma mark - CLLocationManagerDelegate

- (void)locationManager:(CLLocationManager *)manager didUpdateLocations:(NSArray<CLLocation *> *)locations {
    CLLocation * latestLocation = [locations lastObject];
    
    if(latestLocation == nil) {
        return;
    }
    
    SpacedataPassTimeProvider * provider = [SpacedataPassTimeProvider new];
    NSMutableString * endpoint = [NSMutableString stringWithString:[provider endpointString]];
    [endpoint appendFormat:@"?lat=%f&lon=%f", [latestLocation coordinate].latitude, [latestLocation coordinate].longitude];
    
    NSURL * passtimeURL = [NSURL URLWithString:endpoint];
    NSURLRequest * passtimeRequest = [NSURLRequest requestWithURL:passtimeURL];
    [self jsonDataFromApiRequest:passtimeRequest provider:provider completion:^(id jsonObject) {
        NSString * multilineResult = [self risingTimesFromJson:jsonObject];
        dispatch_async(dispatch_get_main_queue(), ^{
            [[[self mapView] userLocation] setTitle:@"Next ISS risings on location:"];
            [self setMultilineRisingTimes:multilineResult];
        });
    }];
}

- (NSString*)risingTimesFromJson:(NSDictionary*)jsonObject
{
    NSMutableString * multilineResult = [NSMutableString string];
    
    NSDateFormatter * defaultFormatter = [NSDateFormatter new];
    [defaultFormatter setLocalizedDateFormatFromTemplate:@"ddmmyyHHMMss"];
    
    for (NSDictionary * risingInfo in jsonObject) {
        NSInteger duration = [[risingInfo objectForKey:@"duration"] integerValue];
        NSInteger unixTime = [[risingInfo objectForKey:@"risetime"] integerValue];
        
        NSDate * risingTime = [NSDate dateWithTimeIntervalSince1970:unixTime*1.0];
        
        NSString * readableTime = [defaultFormatter stringFromDate:risingTime];
        
        if([multilineResult length] > 0) {
            [multilineResult appendString:@"\n"];
        }
        [multilineResult appendFormat:@"%@, %lus", readableTime, duration];
    }
    return multilineResult;
}

- (void)setMultilineRisingTimes:(NSString *)multilineRisingTimes {
    if(_multilineRisingTimes != multilineRisingTimes) {
        _multilineRisingTimes = [multilineRisingTimes copy];
        [self addRisingDetailsToCalloutAccessory];
    }
}

- (void) addRisingDetailsToCalloutAccessory {
    if([[self multilineRisingTimes] length] > 0 && [self risingInfo] != nil) {
        [[self risingInfo] setText:[self multilineRisingTimes]];
    }
}

# pragma mark - MKMapViewDelegate

- (MKAnnotationView*)mapView:(MKMapView*)map
           viewForAnnotation:(nonnull id<MKAnnotation>)annotation
{
    if([annotation isEqual:[self issAnnotation]]) {
        return [self issViewForAnnotation:annotation];
    } else {
        return [self myViewForAnnotiation:annotation];
    }
    return nil;
}

- (MKAnnotationView*)issViewForAnnotation:(id<MKAnnotation>)annotation
{
    MKAnnotationView * existingView = [[self mapView]   dequeueReusableAnnotationViewWithIdentifier:@"issLocationPin"];
    if (existingView != nil) {
        return existingView;
    }

    MKPinAnnotationView * issView = [[MKPinAnnotationView alloc] initWithAnnotation:annotation reuseIdentifier:@"issLocationPin"];
    [issView setCanShowCallout:YES];
    
    if([issView detailCalloutAccessoryView] == nil) {
        UILabel * namesLabel = [self peopleInfo];
        if(namesLabel == nil) {
            [self setPeopleInfo:[UILabel new]];
            namesLabel = [self peopleInfo];
            [namesLabel setLineBreakMode:NSLineBreakByWordWrapping];
            [namesLabel setNumberOfLines:12];
        }
        [issView setDetailCalloutAccessoryView:namesLabel];
        
        NSLayoutConstraint * widthConstraint = [NSLayoutConstraint constraintWithItem:namesLabel attribute:NSLayoutAttributeWidth relatedBy:NSLayoutRelationEqual toItem:nil attribute:NSLayoutAttributeNotAnAttribute multiplier:1.0 constant:320];

        NSLayoutConstraint * heightConstraint = [NSLayoutConstraint constraintWithItem:namesLabel attribute:NSLayoutAttributeHeight relatedBy:NSLayoutRelationEqual toItem:nil attribute:NSLayoutAttributeNotAnAttribute multiplier:1.0 constant:240];

        [namesLabel addConstraint:widthConstraint];
        [namesLabel addConstraint:heightConstraint];
    }
    
    return issView;
}

- (MKAnnotationView*)myViewForAnnotiation:(id<MKAnnotation>)annotation
{
    MKAnnotationView * existingView = [[self mapView]   dequeueReusableAnnotationViewWithIdentifier:@"myLocationView"];
    if (existingView != nil) {
        return existingView;
    }

    MKMarkerAnnotationView * otherView = [[MKMarkerAnnotationView alloc] initWithAnnotation:annotation reuseIdentifier:@"myLocationView"];
    [otherView setUserInteractionEnabled:annotation];
    [otherView setHighlighted:NO];
    [otherView setCanShowCallout:YES];
    [otherView setMarkerTintColor:[UIColor systemBlueColor]];

    if([otherView detailCalloutAccessoryView] == nil) {
        UILabel * risingsLabel = [self risingInfo];
        
        if(risingsLabel == nil) {
            [self setRisingInfo:[UILabel new]];
            risingsLabel = [self risingInfo];
            [risingsLabel setLineBreakMode:NSLineBreakByWordWrapping];
            [risingsLabel setNumberOfLines:5];
            [risingsLabel setText:[self multilineRisingTimes]];
        }
        [otherView setDetailCalloutAccessoryView:risingsLabel];
        
        NSLayoutConstraint * widthConstraint = [NSLayoutConstraint constraintWithItem:risingsLabel attribute:NSLayoutAttributeWidth relatedBy:NSLayoutRelationEqual toItem:nil attribute:NSLayoutAttributeNotAnAttribute multiplier:1.0 constant:240];
        
        NSLayoutConstraint * heightConstraint = [NSLayoutConstraint constraintWithItem:risingsLabel attribute:NSLayoutAttributeHeight relatedBy:NSLayoutRelationEqual toItem:nil attribute:NSLayoutAttributeNotAnAttribute multiplier:1.0 constant:120];
        
        [risingsLabel addConstraint:widthConstraint];
        [risingsLabel addConstraint:heightConstraint];
        
        [self setRisingInfo:risingsLabel];
    }
    return otherView;
}

- (MKOverlayRenderer *)mapView:(MKMapView *)mapView rendererForOverlay:(id<MKOverlay>)overlay {
    
    CLLocationCoordinate2D coordinates[[(MKPolyline*)overlay pointCount]];
    [(MKPolyline*)overlay getCoordinates:coordinates range:NSMakeRange(0, [(MKPolyline*)overlay pointCount])];
    
    MKPolylineRenderer * renderer = [[MKPolylineRenderer alloc] initWithOverlay:overlay];
    [renderer setStrokeColor:[UIColor systemGrayColor]];
    [renderer setLineWidth:1.5];
    return renderer;
}

@end
